//----------------------------------------------------------------------------

//  LCD

//----------------------------------------------------------------------------

#include "Main.h"
#include "Lcd133x64.h"
#include "Font_5x7.h"
#include "Font_8x13.h"

//--------------------------- : -------------------------------------

#define F_I2C  400 // I2C, kbit

#define OFFSET_X 3 //  X,  21 , x < 128 (. 3)
#define OFFSET_Y 3 //  Y,  6  (. 3)

#define FONT_X 5   //  
#define FONT_Y 7   //  

#define HFONT_X 8  //  
#define HFONT_Y 13 //  

// :

#define RES_X 132
#define RES_Y 64

//   :

#define MAX_LIN 6
#define MAX_POS 21

//  :

#define MAX_X 133 //   TIC151
#define MAX_Y (RES_Y / 8)

//----------------------------- UC1601s: -------------------------------------

//UC1601s Address:

#define LCD_ADDR_C    0x70 //control address
#define LCD_ADDR_D    0x72 //data address

#define LCD_SET_COL_L 0x00 //set column address LSB, CA[3:0]
#define LCD_SET_COL_H 0x10 //set column address MSB, CA[7:4]
#define LCD_SET_TC    0x24 //set temperature compensation, TC[1:0]
#define LCD_SET_POWER 0x28 //set power control, PC[2:0]
#define LCD_SET_ADV   0x30 //set advanced program control, APC[R][7:0]
#define LCD_SET_SL    0x40 //set scroll line, SL[5:0]
#define LCD_SET_PAGE  0xB0 //set page address, PA[3:0]
#define LCD_SET_BIAS  0x81 //set Vbias, PM[7:0]
#define LCD_SET_PART  0x84 //set partial display control, LC[4]
#define LCD_SET_RAMA  0x88 //set RAM address control, AC[2:0]
#define LCD_SET_FR    0xA0 //set frame rate, LC[3]
#define LCD_SET_PON   0xA4 //set all pixel on, DC[1]
#define LCD_SET_INV   0xA6 //set inverse display, DC[0]
#define LCD_SET_EN    0xAE //set display enable, DC[2]
#define LCD_SET_MAP   0xC0 //set LCD mapping control, LC[2:1]
#define LCD_RESET     0xE2 //system reset
#define LCD_NOP       0xE3 //NOP
#define LCD_SET_TEST  0xE4 //set test control
#define LCD_SET_BR    0xE8 //set LCD bias ratio, BR[1:0]
#define LCD_SET_CEN   0xF1 //set COM end, CEN[6:0]
#define LCD_SET_DST   0xF2 //set partial display start, DST[6:0]
#define LCD_SET_DEN   0xF3 //set partial display end, DEN[6:0]

//TC[1:0]:

#define LCD_TC0       0x00 //TC[1:0] = 0: temp. coeff.  -0.05%/C (default)
#define LCD_TC1       0x01 //TC[1:0] = 1: temp. coeff.  -0.10%/C
#define LCD_TC2       0x02 //TC[1:0] = 2: temp. coeff.  -0.15%/C
#define LCD_TC3       0x03 //TC[1:0] = 3: temp. coeff.  -0.00%/C

//PC[2:0]

#define LCD_15N       0x00 //PC[0] = 0, panel loading < 15 nF (default)
#define LCD_24N       0x01 //PC[0] = 1, panel loading 15 - 24 nF
#define LCD_EXT_VLCD  0x00 //PC[2:1] = 00, external Vlcd
#define LCD_INT_VLCD  0x06 //PC[2:1] = 11, internal Vlcd (x7) (default)

//LC[4:1]:

#define LCD_MX        0x02 //LC[1] = 1, X mirror
#define LCD_MY        0x04 //LC[2] = 1, Y mirror
#define LCD_80FPS     0x00 //LC[3] = 0, 80 frames per second (default)
#define LCD_100FPS    0x01 //LC[3] = 1, 100 frames per second
#define LCD_DIS_PD    0x00 //LC[4] = 0, disable partial display,
                           //mux rate = CEN + 1 (DST, DEN not used)
#define LCD_ENA_PD    0x01 //LC[4] = 1, enable partial display,
                           //mux rate = DEN - DST + 1

//AC[2:0]:

#define LCD_WA        0x01 //AC[0] = 1, wrap around
#define LCD_AINC      0x02 //AC[1] = 1, auto increment
#define LCD_PID       0x04 //AC[2] = 1, page auto increment direction

//DC[2:0]:

#define LCD_INVERSE   0x01 //DC[0] = 1, inverse display
#define LCD_ALL_ON    0x01 //DC[1] = 1, all pixels on
#define LCD_ENABLE    0x01 //DC[2] = 1, enable display

//BR[1:0]:

#define LCD_BR_6      0x00 //BR[1:0] = 0: bias ratio = 6
#define LCD_BR_7      0x01 //BR[1:0] = 1: bias ratio = 7
#define LCD_BR_8      0x02 //BR[1:0] = 2: bias ratio = 8
#define LCD_BR_9      0x03 //BR[1:0] = 3: bias ratio = 9 (default)

//--------------------------- : ------------------------------------

static char Screen[MAX_X][MAX_Y]; // 
static char MinX, MaxX;     //   X
static char MinY, MaxY;     //   Y
static bool Inv;            // 
static char Px;             // X-
static char Py;             // Y-

//--------------------------  : ------------------------------

void I2C_Start(void);      //  ""  I2C
void I2C_Write(char data); //   I2C
void I2C_Stop(void);       //  ""  I2C

//----------------------------------------------------------------------------
//-------------------------   : ----------------------------
//----------------------------------------------------------------------------

//--------------------------  LCD: ------------------------------

void LCD_Init(void)
{
  Port_RES_0;                         // RESET
  TWBR = (int)(F_CLK * 1E3 / F_I2C / 2 - 8); // I2C
  TWSR = 0;
  LCD_Clear();                        //  
  Port_RES_1;                         // RESET

  I2C_Start();
  I2C_Write(LCD_ADDR_C);              //control address
  I2C_Write(LCD_RESET);               //system reset
  I2C_Stop();

  Delay_ms(10);
  LCD_SetContr(NOM_CONTR);            // 
  LCD_Update();                       //  
  
  I2C_Start();
  I2C_Write(LCD_ADDR_C);              //control address
  //I2C_Write(LCD_SET_MAP + LCD_MY + LCD_MX); //rotate 180 deg.
  I2C_Write(LCD_SET_EN + LCD_ENABLE); //display enable
  I2C_Stop();
}

//-----------------------   LCD: ---------------------------

void LCD_SetContr(char c)
{
  I2C_Start();
  I2C_Write(LCD_ADDR_C);   //control address
  I2C_Write(LCD_SET_BIAS); //set Vbias
  I2C_Write(c);
  I2C_Stop();
}

//-------------------------  RAM LCD: ------------------------------

void LCD_Update(void)
{
  for(char y = MinY; y < MaxY; y++)
  {
    I2C_Start();
    I2C_Write(LCD_ADDR_C);                           //control address
    I2C_Write(LCD_SET_PAGE + MAX_Y - 1 - y);         //Y = y
    I2C_Write(LCD_SET_COL_L + (MinX & 0x0F));        //X = MinX
    I2C_Write(LCD_SET_COL_H + ((MinX >> 4) & 0x0F)); //X = MinX
    I2C_Stop();
    
    I2C_Start();
    I2C_Write(LCD_ADDR_D);                           //data address
    for(char x = MinX; x < MaxX && x < RES_X; x++)
    {
      char c = Screen[x][y];  
      I2C_Write(c);                                  //write data
    }
    I2C_Stop();
  }
  MinX = MAX_X; MinY = MAX_Y;
  MaxX = 0; MaxY = 0;
}

//--------------------   ""  I2C: ---------------------

void I2C_Start(void)
{
  TWCR = (1 << TWINT) | (1 << TWSTA) | (1 << TWEN);
  while(!(TWCR & (1 << TWINT)));
}

//-------------------------    I2C: -----------------------------

void I2C_Write(char data)
{
  TWDR = data;
  TWCR = (1 << TWINT) | (1 << TWEN);
  while(!(TWCR & (1 << TWINT)));
}

//--------------------   ""  I2C: ---------------------

void I2C_Stop(void)
{
  TWCR = (1 << TWINT) | (1 << TWSTO) | (1 << TWEN);
}

//----------------------------------------------------------------------------
//-----------------------    : -------------------------
//----------------------------------------------------------------------------

//------------------------   : --------------------------

void LCD_Clear(void)
{
  for(char y = 0; y < MAX_Y; y++)
    for(char x = 0; x < MAX_X; x++)
      Screen[x][y] = 0;
  Px = OFFSET_X;
  Py = OFFSET_Y;
  MinX = 0; MinY = 0;
  MaxX = MAX_X; MaxY = MAX_Y;
}

//------------------     : --------------------

void LCD_Picture(char __flash *p)
{
  for(char y = 0; y < MAX_Y; y++)
    for(char x = 0; x < MAX_X; x++)
      Screen[x][y] = *p++;
  MinX = 0; MinY = 0;
  MaxX = MAX_X; MaxY = MAX_Y;
}

//-------------------------  : -----------------------------

void LCD_SetXY(char x, char y)
{
  Px = x;
  Py = y;
}

//--------------------------   x: ----------------------------

char LCD_GetX(void)
{
  return(Px);
}

//--------------------------   y: ----------------------------

char LCD_GetY(void)
{
  return(Py);
}

//----------------------    : -------------------------

void LCD_SetLinPos(char lin, char pos)
{
  Px = (pos - 1) * (FONT_X + 1) + OFFSET_X; //   X
  Py = (lin - 1) * (FONT_Y + 3) + OFFSET_Y; //   Y
}

//------------------------   : ---------------------------

char LCD_GetPos(void)
{
  return((Px - OFFSET_X) / (FONT_X + 1) + 1);
}

//-------------------- / : ------------------------

void LCD_Inv(bool inv)
{
  Inv = inv;
}

//--------      (  ): ---------

void LCD_PutChar(char ch)
{
  if(ch == '\n') // 
  {
    Px = OFFSET_X;
    Py = Py + (FONT_Y + 3);
    return;
  }    
  if(Px > RES_X - (FONT_X + 2)) return; //   X
  if(Py > RES_Y - (FONT_Y + 2)) return; //   Y
  char a = 7 - (Py >> 3);    //  
  char z = Py & 7;           // 
  if(Inv) //   :
  {
    unsigned int c = ((1 << (FONT_Y + 2)) - 1) << z;
    char m1 = (char)c;
    char m2 = (char)(c >> 8);
    Screen[Px][a] |= m1;
    Screen[Px][a - 1] |= m2;
    Screen[Px + (FONT_X + 1)][a] |= m1;
    Screen[Px + (FONT_X + 1)][a - 1] |= m2;
    for(char i = 0; i < FONT_X; i++)
    {
      c = (unsigned int)Font[ch][i] << (z + 1);
      char xi = Px + i + 1;
      Screen[xi][a] = (Screen[xi][a] | m1) & ~((char)c);
      Screen[xi][a - 1] = (Screen[xi][a - 1] | m2) & ~((char)(c >> 8));
    }
  }
  else //   :
  {
    unsigned int c = ((1 << (FONT_Y + 2)) - 1) << z;
    char m1 = ~(char)c;
    char m2 = ~(char)(c >> 8);
    Screen[Px][a] &= m1;
    Screen[Px][a - 1] &= m2;
    Screen[Px + (FONT_X + 1)][a] &= m1;
    Screen[Px + (FONT_X + 1)][a - 1] &= m2;
    for(char i = 0; i < FONT_X; i++)
    {
      c = (unsigned int)Font[ch][i] << (z + 1);
      char xi = Px + i + 1;
      Screen[xi][a] = Screen[xi][a] & m1 | ((char)c);
      Screen[xi][a - 1] = Screen[xi][a - 1] & m2 | ((char)(c >> 8));
    }
  }
  //   :
  if((Px) < MinX) MinX = Px;
  if((Px + (FONT_X + 2)) > MaxX) MaxX = Px + (FONT_X + 2);
  if((a - 1) < MinY) MinY = a - 1;
  if((a + 1) > MaxY) MaxY = a + 1; 
  // :
  Px = Px + (FONT_X + 1);
}

//------------  null-terminated string (  ): -------------

void LCD_PutString(char __flash *s)
{
  while(*s) LCD_PutChar(*s++);
}

//-------  H-    (  ): --------
  
void LCD_PutHChar(char ch)
{
  switch(ch)
  {
  case '.': ch = 10; break;
  case '-': ch = 11; break; 
  case 'S': ch = 12; break;
  case 'G': ch = 13; break; 
  case ' ': ch = 14; break;
  default: if(ch >= 0x30 && ch <= 0x39) ch -= 0x30; else ch = 11;
  }
  if(Px > RES_X - (HFONT_X + 2)) return; //   X
  if(Py > RES_Y - (HFONT_Y + 2)) return; //   Y
  char a = 7 - (Py >> 3);     //  
  char z = Py & 7;            // 
  if(Inv) //   :
  {
    long c = ((1L << (HFONT_Y + 2)) - 1) << z;
    char m1 = (char)c;
    char m2 = (char)(c >> 8);
    char m3 = (char)(c >> 16);
    Screen[Px][a] |= m1;
    Screen[Px][a - 1] |= m2;
    Screen[Px][a - 2] |= m3;
    Screen[Px + (HFONT_X + 1)][a] |= m1;
    Screen[Px + (HFONT_X + 1)][a - 1] |= m2;
    Screen[Px + (HFONT_X + 1)][a - 2] |= m3;
    for(char i = 0; i < HFONT_X; i++)
    {
      char xi = Px + i + 1;
      c = (long)((long)(FontH[ch][i] << 8) | FontH[ch][i + HFONT_X]) << (z + 1);
      Screen[xi][a] = (Screen[xi][a] | m1) & ~((char)c);
      Screen[xi][a - 1] = (Screen[xi][a - 1] | m2) & ~((char)(c >> 8));
      Screen[xi][a - 2] = (Screen[xi][a - 2] | m3) & ~((char)(c >> 16));
    }
  }
  else //   :
  {
    long c = ((1L << (HFONT_Y + 2)) - 1) << z;
    char m1 = ~(char)c;
    char m2 = ~(char)(c >> 8);
    char m3 = ~(char)(c >> 16);
    Screen[Px][a] &= m1;
    Screen[Px][a - 1] &= m2;
    Screen[Px][a - 2] &= m3;
    Screen[Px + (HFONT_X + 1)][a] &= m1;
    Screen[Px + (HFONT_X + 1)][a - 1] &= m2;
    Screen[Px + (HFONT_X + 1)][a - 1] &= m3;
    for(char i = 0; i < HFONT_X; i++)
    {
      char xi = Px + i + 1;
      long c = (long)((long)(FontH[ch][i] << 8) | FontH[ch][i + HFONT_X]) << (z + 1);
      Screen[xi][a] = Screen[xi][a] & m1 | ((char)c);
      Screen[xi][a - 1] = Screen[xi][a - 1] & m2 | ((char)(c >> 8));
      Screen[xi][a - 2] = Screen[xi][a - 2] & m3 | ((char)(c >> 16));
    }
  }
  //   :
  if((Px) < MinX) MinX = Px;
  if((Px + (HFONT_X + 2)) > MaxX) MaxX = Px + (HFONT_X + 2);
  if((a - 2) < MinY) MinY = a - 2;
  if((a + 1) > MaxY) MaxY = a + 1; 
  // :
  Px = Px + (HFONT_X + 2);
}

//-----------  null-terminated H-string (  ): ------------

void LCD_PutHString(char __flash *s)
{
  while(*s) LCD_PutHChar(*s++);
}

//----------------------------------------------------------------------------
