;------------------------------- DDS2.asm -----------------------------------

;Software 2-channel DDS implementation module
;using I2S audio DAC TDA1543.
;I2S is emulated with hardware SPI plus OC signal.

;Used ports:

;I2SDATA - SPI MOSI
;I2SBCK - SPI SCK
;I2SWS - OC1B

;Connections:

;I2SDATA -> NOT gate -> DAC DATA
;I2SBCK -> NOT gate -> DAC BCK
;I2SWS -> NOT gate -> DAC WS

;----------------------------- Constantes: ----------------------------------

.equ	FDIV    = 97            ;DDS clock divider

;----------------------------- Variables: -----------------------------------

.DSEG	;data segment (internal SRAM)

FreeMem:

.org	(FreeMem + 0x100) & 0xFF00 ;align table to page

LUT:	.byte 514		;LUT

;------------------------------ Macros: -------------------------------------

;Stop DDS:

.macro	StopDDS
	sts	TIMSK1,zero	;disable output compare interrupt
.endm

;Start DDS:

.macro	StartDDS
        ldi     temp,FDIV - 14
	sts	TCNT1H,zero     ;load non-zero value to
	sts	TCNT1L,temp     ;reduce phase jump

	ldi	temp,(1<<COM1B1)
	sts	TCCR1A,temp	;OC1B clear on compare
	ldi	temp,(1<<FOC1B)
	sts	TCCR1C,temp	;force OC1B
	ldi	temp,(1<<COM1B0)
	sts	TCCR1A,temp	;OC1B toggle on compare

	ldi	temp,(1<<OCF1A)	
	out	TIFR1,temp	;clear pending timer interrupt
	sts	TIMSK1,temp	;enable output compare interrupt
.endm

;-------------------------------- Code: -------------------------------------

.CSEG	;Code segment

;Timer1 output compare A interrupt (DDS implementation)
;To minimize execution time handle must be located
;direct at vector address. It eliminates RJMP command.

;AFreqK,L,M,N, BFreqK,L,M,N - current frequency (register variables)
;LPhaseK,L,M,N,P, RPhaseK,L,M,N,P - current phase (register variables)
;SinL,H - sin samples magnitude (register variables)
;ZL,ZH - data pointer
;tsreg - SREG store
;r0,r1 - used with mul instruction

                                ;4      address store
DDS:    lds     ZL,TCNT1L	;2      TCNT1L = 5..8
        ldi     ZH,high(JmpTab) ;1
        ijmp                    ;2

.org	(PC + 0x100) & 0xFF00   ;align to page

JmpTab: nop                     ;       dummy
        nop                     ;       dummy
        nop                     ;       dummy
        nop                     ;       dummy
        nop                     ;       dummy
        nop                     ;0/1    TCNT1L = 5
        nop                     ;0/1    TCNT1L = 6
        nop                     ;0/1    TCNT1L = 7
                                ;       TCNT1L = 8
	in	tsreg,SREG	;10/13  save status register

	out	SPDR,SinH	;1 ---> load DAC L high byte

	add	APhaseK,AFreqK	;1      Phase(0..33) = Phase(0..33) + Freq(0..31)
	adc	APhaseL,AFreqL	;1
	adc	APhaseM,AFreqM  ;1
	adc	APhaseN,AFreqN  ;1
	adc	APhaseP,zero    ;1

	mov	ZL,APhaseN	;1      ZL = wa (word address)
	mov	r0,APhaseM	;1      r0 = dx
	sbrc	APhaseP,0	;2/1    if(Phase.24 == 0)
	com	ZL		;0/1    wa = !wa
	sbrc	APhaseP,0	;2/1    if(Phase.24 == 0)
	com	r0		;0/1    dx = !dx

	ldi	ZH,high(LUT)	;1      ZH = table base (low(LUT) = 0)
	add	ZL,ZL		;1      offset * 2 (word offset)
	adc	ZH,zero         ;1

	ldi	SinH,46         ;1
	sts	OCR1BL,SinH     ;2

	out	SPDR,SinL	;19 ---> load DAC L low byte

	ld	SinL,Z+		;2      SinL = lo sin[x]
	ld	SinH,Z+		;2      SinH = hi sin[x]
	ld	r1,Z		;2      r1 = lo sin[x + 1]
	sub	r1,SinL		;1      r1 = dA
	mul	r1,r0		;2      r1,r0 = dA * dx
	rol	r0		;1      C = 1 if r0.7 == 1
	adc	SinL,r1		;1      SinH:SinL = sin[x] + round(r1:r0 / 256)
	adc	SinH,zero	;1      SinH:SinL = A

	sbrs	APhaseP,1       ;2/1	
	rjmp	ph_cd1		;0/2    jump if Phase.25 == 0

ph_ab1:	com	SinL		;1/0    SIN > 0, data line has NOT gate,
	com	SinH		;1/0    SinH:SinL = !SinH:SinL
	rjmp	ph_ad1          ;2/0
	
ph_cd1:	sec                     ;0/1
	sbc	SinL,zero       ;0/1    SIN < 0, data line has NOT gate,
	sbc	SinH,zero	;0/1    SinH:SinL = SinH:SinL - 1

ph_ad1:	out	SPDR,SinH	;38 ---> load DAC R high byte
	
	add	BPhaseK,BFreqK	;1      Phase(0..33) = Phase(0..33) + Freq(0..31)
	adc	BPhaseL,BFreqL	;1
	adc	BPhaseM,BFreqM  ;1
	adc	BPhaseN,BFreqN  ;1
	adc	BPhaseP,zero    ;1      

	mov	ZL,BPhaseN	;1      ZL = wa (word address)
	mov	r0,BPhaseM	;1      r0 = dx
	sbrc	BPhaseP,0	;2/1    if(Phase.24 == 0)
	com	ZL		;0/1    wa = !wa
	sbrc	BPhaseP,0	;2/1    if(Phase.24 == 0)
	com	r0		;0/1    dx = !dx

	ldi	ZH,high(LUT)	;1      ZH = table base (low(LUT) = 0)
	add	ZL,ZL		;1      offset * 2 (word offset)
	adc	ZH,zero         ;1      

	ldi	SinH,83         ;1
	sts	OCR1BL,SinH     ;2
	
	out	SPDR,SinL	;56 ---> load DAC R low byte
	
	ld	SinL,Z+		;2      SinL = lo sin[x]
	ld	SinH,Z+		;2      SinH = hi sin[x]
	ld	r1,Z		;2      r1 = lo sin[x + 1]
	sub	r1,SinL		;1      r1 = dA
	mul	r1,r0		;2      r1,r0 = dA * dx
	rol	r0		;1      C = 1 if r0.7 == 1
	adc	SinL,r1		;1      SinH:SinL = sin[x] + round(r1:r0 / 256)
	adc	SinH,zero	;1      SinH:SinL = A

	sbrs	BPhaseP,1	;2/1	
	rjmp	ph_cd2		;0/2    jump if Phase.25 == 0

ph_ab2:	com	SinL		;1/0    SIN > 0, data line has NOT gate,
	com	SinH		;1/0    SinH:SinL = !SinH:SinL
	rjmp	ph_ad2          ;2/0
	
ph_cd2:	sec                     ;0/1
	sbc	SinL,zero	;0/1    SIN < 0, data line has NOT gate,
	sbc	SinH,zero	;0/1    SinH:SinL = SinH:SinL - 1

ph_ad2:	out	SREG,tsreg	;1      restore status register
	reti                    ;4
                                ;89/92  total
	
;----------------------------------------------------------------------------

;Init DDS subsystem:

iDDS:

;SPI setup:

	ldi	temp,(1<<SPE) | (1<<MSTR) | (1<<CPOL)	
	out	SPCR,temp	;SPI enable, MSB first, master

	ldi	temp,(1<<SPI2X)	
	out	SPSR,temp	;double SPI speed

;Build sin LUT in RAM:

	table	SinTab
	ldy	LUT
SCopy:	lpm	temp,Z+
	st	Y+,temp
	cpi	YH,high(LUT + 514)
	brne	SCopy
	cpi	YL, low(LUT + 514)
	brne	SCopy

;Build triangle LUT in RAM:

;	ldz	0
;	ldy	LUT
;TCopy:	st	Y+,ZL
;	st	Y+,ZH
;	subi	ZL, low(-127)
;	sbci	ZH,high(-127)
;	cpi	YH,high(LUT + 514)
;	brne	TCopy
;	cpi	YL, low(LUT + 514)
;	brne	TCopy

;Variables init

	clr	SinL		;clear sin amplitude
	clr	SinH

	clr	APhaseK         ;clear phase for channel A
	clr	APhaseL
	clr	APhaseM
	clr	APhaseN
	clr	APhaseP

	clr	AFreqK          ;clear frequency for channel A
	clr	AFreqL
	clr	AFreqM
	clr	AFreqN

	clr	BPhaseK         ;clear phase for channel B
	clr	BPhaseL
	clr	BPhaseM
	clr	BPhaseN
	clr	BPhaseP

	clr	BFreqK          ;clear frequency for channel B
	clr	BFreqL
	clr	BFreqM
	clr	BFreqN

;	ldi	temp,0xf3       ;channel A: 5 KHz
;	mov	AFreqK,temp
;	ldi	temp,0xfd
;	mov	AFreqL,temp
;	ldi	temp,0xd4
;	mov	AFreqM,temp
;	ldi	temp,0x18
;	mov	AFreqN,temp

;	ldi	temp,0x85       ;channel B: 50 KHz
;	mov	BFreqK,temp
;	ldi	temp,0xeb
;	mov	BFreqL,temp
;	ldi	temp,0x51
;	mov	BFreqM,temp
;	ldi	temp,0xf8
;	mov	BFreqN,temp

;Timer 1 setup:

	ldi	temp,high(FDIV - 1)
	sts	OCR1AH,temp	
	ldi	temp,low(FDIV - 1)
	sts	OCR1AL,temp

	ldi	temp,(1<<WGM12) | (1<<CS10)	
	sts	TCCR1B,temp	;clear on compare match, CK/1

	sts	OCR1BH,zero	
	ldi	temp,44
	sts	OCR1BL,temp

        StartDDS
	ret

;----------------------------------------------------------------------------

SinTab:

.include "sin256.asm"		;sine wave table

;----------------------------------------------------------------------------
