;------------------------------- Port.asm -----------------------------------

;Co-processor communication module using UART.

;Connections:

;Co-processor RXD -> Processor TXD
;Co-processor TXD -> Processor RXD
;Co-processor INT -> Processor INT

;----------------------------- Constantes: ----------------------------------

.equ	Baud	= 1250000	;baud rate, bps

.equ UBRRV = (((F_CLK * 10) / (16 * Baud)) - 5) / 10
.if  UBRRV > MAXWORD
.error "out of range constant"
.endif

.equ	C_SET_F	= 0x01	        ;SET FREQUENCY command
.equ	C_SET_P	= 0x02	        ;SET PHASE command
.equ	C_SET_D	= 0x03	        ;SET DELTA PHASE command
.equ	C_SET_S	= 0x04	        ;SET SHAPE command

.equ    SH_S    = 0x00          ;sine shape
.equ	SH_M    = 0x01          ;square shape

.equ	CHAN_A	= 4	        ;bit 4 - channel A flag
.equ	CHAN_B	= 5	        ;bit 5 - channel B flag

;----------------------------- Variables: -----------------------------------

.DSEG	;data segment

Comm:	.byte	1		;command code
Buff:	.byte	4		;RX buffer
Bpnt:   .byte   1               ;data pointer

;-------------------------------- Code: -------------------------------------

.CSEG	;Code segment

;----------------------------------------------------------------------------

;Init port:

iPort:  sts     Comm,zero       ;clear command code

;UART0 setup (9 bit):

	sts	UCSR0A,zero
	ldi	temp,(1<<RXEN0) | (1<<TXEN0) | (1<<UCSZ02)
	sts	UCSR0B,temp
	ldi	temp,(1<<UCSZ01) | (1<<UCSZ00)
	sts	UCSR0C,temp
	ldi	temp,high(UBRRV)
	sts	UBRR0H,temp
	ldi	temp,low(UBRRV)
	sts	UBRR0L,temp
	ret

;----------------------------------------------------------------------------

;RX frame:

ePort:  lds     cnt,UCSR0A      ;read status register
        bbrc    cnt,RXC0,exi    ;exit if no received byte
        lds     cnt,UCSR0B      ;read RXB8 bit
        lds     temp,UDR0       ;read data byte
        Port_CINT_CPL           ;toggle CINT signal
        bbrc    cnt,RXB80,dat   ;if RXB8 == 1 then command

;Command byte received:

cmd:    sts     Comm,temp       ;save command code
        sts     Bpnt,zero       ;clear data pointer
        rjmp    exi

;Data byte received:

dat:    lds     cnt,Comm        ;load command code
        tst     cnt             ;check code
        breq    exi             ;skip unknown data bytes

        lds     cnt,Bpnt        ;load data pointer
        ldx     Buff            ;load buffer base
        add     XL,cnt          ;add offset
        adc     XH,zero
        st      X,temp          ;save data byte
        inc     cnt             ;advance data pointer
        sts     Bpnt,cnt        ;save data pointer
        cpi     cnt,4           ;check pointer
        brne    exi             ;frame not complete

;Execute command:

        lds     temp,Comm       ;load command code
        andi    temp,0x0f       ;select command bits
        cpi     temp,C_SET_F    ;SET FREQUENCY command
        brne    cmd1
        rjmp    SetF
cmd1:   cpi     temp,C_SET_P    ;SET PHASE command
        brne    cmd2
        rjmp    SetP
cmd2:   cpi     temp,C_SET_D    ;SET DELTA PHASE command
        brne    cmd3
        rjmp    SetD
cmd3:   cpi     temp,C_SET_S    ;SET SHAPE command
        brne    cmd4
        rjmp    SetS

cmd4:   sts     Comm,zero       ;clear command code
exi:    ret

;----------------------------------------------------------------------------

;SET FREQUENCY:

SetF:   lds     temp,Comm       ;load command code
        StopDDS
        bbrc    temp,CHAN_A,f_b

        ldx     Buff            ;load buffer base
        ld      AFreqK,X+       ;set frequency for channel A
        ld      AFreqL,X+
        ld      AFreqM,X+
        ld      AFreqN,X+

f_b:    bbrc    temp,CHAN_B,f_n

        ldx     Buff            ;load buffer base
        ld      BFreqK,X+       ;set frequency for channel B
        ld      BFreqL,X+
        ld      BFreqM,X+
        ld      BFreqN,X+

f_n:    StartDDS
        sts     Comm,zero       ;clear command code
        ret

;----------------------------------------------------------------------------

;SET PHASE:

SetP:   lds     temp,Comm       ;load command code
        StopDDS
        bbrc    temp,CHAN_A,p_b

        ldx     Buff            ;load buffer base
        clr     APhaseK         ;set phase for channel A
        ld      APhaseL,X+
        ld      APhaseM,X+
        ld      APhaseN,X+
        ld      APhaseP,X+

p_b:    bbrc    temp,CHAN_B,p_n

        ldx     Buff            ;load buffer base
        clr     BPhaseK         ;set phase for channel B
        ld      BPhaseL,X+
        ld      BPhaseM,X+
        ld      BPhaseN,X+
        ld      BPhaseP,X+

p_n:    StartDDS
        sts     Comm,zero       ;clear command code
        ret

;----------------------------------------------------------------------------

;SET DELTA PHASE:

SetD:   StopDDS
        lds     temp,Comm       ;load command code
        bbrc    temp,CHAN_A,d_b

        ldx     Buff            ;load buffer base
        ld      cnt,X+          ;set delta phase for channel A
        add     APhaseL,cnt
        ld      cnt,X+
        adc     APhaseM,cnt
        ld      cnt,X+
        adc     APhaseN,cnt
        ld      cnt,X+
        adc     APhaseP,cnt

d_b:    bbrc    temp,CHAN_B,d_n

        ldx     Buff            ;load buffer base
        ld      cnt,X+          ;set delta phase for channel B
        add     BPhaseL,cnt
        ld      cnt,X+
        adc     BPhaseM,cnt
        ld      cnt,X+
        adc     BPhaseN,cnt
        ld      cnt,X+
        adc     BPhaseP,cnt

d_n:    StartDDS
        sts     Comm,zero       ;clear command code
        ret

;----------------------------------------------------------------------------

;SET SHAPE:

SetS:   lds     temp,Comm       ;load command code
        lds     cnt,Buff        ;load command data
        bbrc    temp,CHAN_A,s_b

        cpi     cnt,SH_M
        breq    s1_a
s0_a:   Port_SQRA_0
        rjmp    s_b    
s1_a:   Port_SQRA_1

s_b:    bbrc    temp,CHAN_B,s_n

        cpi     cnt,SH_M
        breq    s1_b
s0_b:   Port_SQRB_0
        rjmp    s_n    
s1_b:   Port_SQRB_1
        
s_n:    sts     Comm,zero       ;clear command code
        ret

;----------------------------------------------------------------------------
